//  pcanbasic_jni.cpp
//
//  ~~~~~~~~~~~~
//
//  PCANBasic JNI Interface
//
//  ~~~~~~~~~~~~
//
//  ------------------------------------------------------------------
//  Author : Jonathan Urban
//  Co-Author : Thomas Haber (thomas@toem.de) for Linux port
//  Last change: 08.01.2013
//
//  Language: C++
//  ------------------------------------------------------------------
//
//  Copyright (C) 1999-2010  PEAK-System Technik GmbH, Darmstadt
//  more Info at http://www.peak-system.com
//  ------------------------------------------------------------------
//
// linux@peak-system.com
// www.peak-system.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <asm/types.h>

#include "peak_can_basic_PCANBasic.h"

#define DWORD  __u32
#define WORD   unsigned short
#define BYTE   unsigned char
#define LPSTR  char*

#define MAX_PATH 256

#include "PCANBasic.h"
#include "pcanbasic_jni.h"

#define true 1
#define false 0

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Jni on load. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="jvm">		[in] If non-null, the jvm. </param>
/// <param name="reserved">	[in]reserved. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved) {
	return JNI_VERSION_1_2;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Jni on unload. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="jvm">		[in] If non-null, the jvm. </param>
/// <param name="reserved">	[in] reserved. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved) {
	return;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Initializes the PCANBasic API. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">	[in] If non-null, the env. </param>
/// <param name="obj">	The object. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jboolean JNICALL Java_peak_can_basic_PCANBasic_initializeAPI(
		JNIEnv *env, jobject obj) {
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Get Java Enum Value By Class Path and Constant Name. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">				[in] If non-null, the env. </param>
/// <param name="target">			[out] If non-null, the target. </param>
/// <param name="className">		Name of the class. </param>
/// <param name="valueName">		Name of the value. </param>
/// <param name="byteCodeTypeName">	Name of the byte code type. </param>
///
/// <returns>	The class enum value. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

int GetClassEnumValue(JNIEnv *env, jobject* target, const char *className,
		const char *valueName, const char *byteCodeTypeName) {
	jclass cls;
	jfieldID fid;

	// Find JClass
	cls = (*env)->FindClass(env, className);
	if (cls == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"FindClass failed.");
		return -1;
	}

	// Get JField
	fid = (*env)->GetStaticFieldID(env, cls, valueName, byteCodeTypeName);
	if (fid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetStaticFieldID failed.");
		return -1;
	}

	// Get JObject
	*target = (*env)->GetStaticObjectField(env, cls, fid);
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Throws a given exception in JVM. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">	[in] If non-null, the env. </param>
/// <param name="name">	The name. </param>
/// <param name="msg">	The message. </param>
////////////////////////////////////////////////////////////////////////////////////////////////////

void ThrowExByName(JNIEnv *env, const char *name, const char *msg) {
	// Find JClass
	jclass cls = (*env)->FindClass(env, name);
	if (cls != NULL) {
		(*env)->ThrowNew(env, cls, msg);
	}
	// Free Resources
	(*env)->DeleteLocalRef(env, cls);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Initializes a PCAN Channel. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
/// <param name="Btr0Btr1">	"The speed for the communication (BTR0BTR1 code)". </param>
/// <param name="HwType">	"NON PLUG&PLAY: The type of hardware and operation mode". </param>
/// <param name="IOPort">	"NON PLUG&PLAY: The I/O address for the parallel port". </param>
/// <param name="Interupt">	"NON PLUG&PLAY: Interrupt number of the parallel port". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_Initialize(JNIEnv *env,
		jobject obj, jobject Channel, jobject Btr0Btr1, jobject HwType,
		jint IOPort, jshort Interupt) {
	TPCANStatus status;
	jobject jStatus;
	BYTE pChannel;
	int pBtr0Btr1;
	BYTE pHwType;

	// Parse Channel
	pChannel = (BYTE) ParseEnumValueFromJava(env, &Channel, "()B");

	// Parse Btr0Btr1
	pBtr0Btr1 = ParseEnumValueFromJava(env, &Btr0Btr1, "()I");

	// Parse HwType
	pHwType = (BYTE) ParseEnumValueFromJava(env, &HwType, "()B");

	printf("%i,%i,%i,%i,%i", (int) pChannel, (int) pBtr0Btr1, (int) pHwType,
			(int) IOPort, (int) Interupt);

	//Call PCANBasic Function
	status = CAN_Initialize(pChannel, pBtr0Btr1, pHwType, IOPort, Interupt);
	printf("->(%i)", (int) status);

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>
/// Returns a descriptive text of a given TPCANStatus error code, in any desired language.
/// </summary>
///
/// <remarks>
/// The current languages available for translation are: Neutral (0x00), German (0x07), English
/// (0x09), Spanish (0x0A), Italian (0x10) and French (0x0C)
/// </remarks>
///
/// <param name="env">			[in,out] If non-null, the env. </param>
/// <param name="obj">			The object. </param>
/// <param name="Error">		"A TPCANStatus error code". </param>
/// <param name="Language">		"Indicates a 'Primary language ID'". </param>
/// <param name="stringBuffer">	Buffer for string data. </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
///
/// ### <param name="Buffer">	"Buffer for a null terminated char array". </param>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_GetErrorText(
		JNIEnv *env, jobject obj, jobject Error, jshort Language,
		jobject stringBuffer) {
	TPCANStatus status;
	char strInfo[256] = "";
	jobject jStatus;
	int pError;

	// Parse Btr0Btr1
	pError = ParseEnumValueFromJava(env, &Error, "()I");

	//Call PCANBasic Function
	status = CAN_GetErrorText(pError, Language, strInfo);
	printf("%s-%i-%i", strInfo, (int) Language, (int) pError);
	// Append Version Info To String Buffer
	if (status == PCAN_ERROR_OK)
		appendTextToJavaStringBuffer(env, stringBuffer, strInfo);

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Uninitializes one or all PCAN Channels initialized by CAN_Initialize. </summary>
///
/// <remarks>
/// Giving the TPCANHandle value "PCAN_NONEBUS", uninitialize all initialized channels.
/// </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_Uninitialize(
		JNIEnv *env, jobject obj, jobject Channel) {
	TPCANStatus status;
	jobject jStatus;
	BYTE pChannel;

	// Parse Channel
	pChannel = (BYTE) ParseEnumValueFromJava(env, &Channel, "()B");

	//Call PCANBasic Function
	status = CAN_Uninitialize(pChannel);

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Reads a CAN message from the receive queue of a PCAN Channel. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">				[in,out] If non-null, the env. </param>
/// <param name="obj">				The object. </param>
/// <param name="Channel">			"The handle of a PCAN Channel". </param>
/// <param name="MessageBuffer">	"A TPCANMsg structure buffer to store the CAN message". </param>
/// <param name="TimestampBuffer">	"A TPCANTimestamp structure buffer to get the reception time
/// 								of the message. If this value is not desired, this parameter
/// 								should be passed as NULL". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_Read(JNIEnv *env,
		jobject obj, jobject Channel, jobject MessageBuffer,
		jobject TimestampBuffer) {
	BYTE pChannel;
	TPCANMsg mymsg;
	TPCANTimestamp myRcvTime;
	TPCANStatus status;
	jobject jStatus;
	jclass cls;
	jmethodID mid;

	// Parse Channel
	pChannel = (BYTE) ParseEnumValueFromJava(env, &Channel, "()B");

	// Call CAN-API function
	if (TimestampBuffer != NULL)
		status = CAN_Read(pChannel, &mymsg, &myRcvTime);
	else
		status = CAN_Read(pChannel, &mymsg, NULL);

	if (status == PCAN_ERROR_OK) {
		if (TimestampBuffer != NULL) {
			// Update Java rcvTime variable
			cls = (*env)->GetObjectClass(env, TimestampBuffer);
			if (cls == NULL) {
				ThrowExByName(env, "java/lang/NullPointerException",
						"GetObjectClass for TPCANTimestamp failed.");
				ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
				return jStatus;
			}

			// set millis	
			mid = (*env)->GetMethodID(env, cls, "setMillis", "(I)V");
			if (mid == NULL) {
				ThrowExByName(env, "java/lang/NullPointerException",
						"GetMethodID 'setMillis' for TPCANTimestamp failed.");
				ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
				return jStatus;
			}
			(*env)->CallVoidMethod(env, TimestampBuffer, mid,
					(jint) myRcvTime.millis);

			// set millis_overflow	
			mid = (*env)->GetMethodID(env, cls, "setMillis_overflow", "(S)V");
			if (mid == NULL) {
				ThrowExByName(env, "java/lang/NullPointerException",
						"GetMethodID 'setMillis_overflow' for TPCANTimestamp failed.");
				ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
				return jStatus;
			}
			(*env)->CallVoidMethod(env, TimestampBuffer, mid,
					(jshort) myRcvTime.millis_overflow);

			// set micros	
			mid = (*env)->GetMethodID(env, cls, "setMicros", "(S)V");
			if (mid == NULL) {
				ThrowExByName(env, "java/lang/NullPointerException",
						"GetMethodID 'setMicros' for TPCANTimestamp failed.");
				ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
				return jStatus;
			}
			(*env)->CallVoidMethod(env, TimestampBuffer, mid,
					(jshort) myRcvTime.micros);
		}

		// Update Java message variable
		ParseCanMessageToJava(env, &mymsg, &MessageBuffer);
	}

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>
/// Sets the hadle of the Receive-Event for the Channel. static method
/// peak.can.basic.RcvEventDispatcher.dispatchRcvEvent is used to notify each Receive-Event.
/// </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_SetRcvEvent(
		JNIEnv *env, jobject obj, jobject Channel) {

	jobject jStatus;
	BYTE pChannel;
	CANReadThreadFuncParameter* parameter;

	// Parse Channel
	pChannel = (BYTE) ParseEnumValueFromJava(env, &Channel, "()B");

	// Save pointer to the loaded JVM
	(*env)->GetJavaVM(env, &m_vm);

	//Start Thread
	parameter = malloc(sizeof(CANReadThreadFuncParameter));
	parameter->channel = pChannel;
	parameter->terminating = 0;
	void* CANReadThreadFunc(void*);
	pthread_create(&parameter->thread, NULL, CANReadThreadFunc, parameter);
	threadParameterList[pChannel] = parameter;

	//Parse Result
	ParseTPCANStatusToJava(env, PCAN_ERROR_OK, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>
/// Resets the hadle of the Receive-Event for the Channel. static method
/// peak.can.basic.RcvEventDispatcher.dispatchRcvEvent is used to notify each Receive-Event.
/// </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel. </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_ResetRcvEvent(
		JNIEnv *env, jobject obj, jobject Channel) {
	jobject jStatus;
	BYTE pChannel;
	CANReadThreadFuncParameter* parameter;

	// Parse Channel
	pChannel = (BYTE) ParseEnumValueFromJava(env, &Channel, "()B");

	//Get CANReadThreadFuncParameter
	parameter = threadParameterList[pChannel];

	//Terminate the thread
	parameter->terminating = true;

	//Parse Result
	ParseTPCANStatusToJava(env, PCAN_ERROR_OK, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Retrieves a PCAN Channel value. </summary>
///
/// <remarks>
/// Parameters can be present or not according with the kind of Hardware (PCAN Channel) being
/// used. If a parameter is not available, a PCAN_ERROR_ILLPARAMTYPE error will be returned.
/// </remarks>
///
/// <param name="env">			[in,out] If non-null, the env. </param>
/// <param name="obj">			The object. </param>
/// <param name="Channel">		"The handle of a PCAN Channel". </param>
/// <param name="Parameter">	"The TPCANParameter parameter to get". </param>
/// <param name="Buffer">		"Buffer for the parameter value". </param>
/// <param name="BufferLength">	"Size in bytes of the buffer". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_GetValue(JNIEnv *env,
		jobject obj, jobject Channel, jobject Parameter, jobject Buffer,
		jint BufferLength) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;
	TPCANParameter pParamater;
	jclass cls;
	jmethodID mid;
	char charBuffer[256] = "";
	int intBuffer;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Parse Parameter
	pParamater
			= (TPCANParameter) ParseEnumValueFromJava(env, &Parameter, "()I");

	//Check Buffer Is Instance Of StringBuffer
	if ((*env)->IsInstanceOf(env, Buffer, (*env)->FindClass(env,
			"java/lang/StringBuffer")) == JNI_TRUE) {
		//Call PCANBasic Function
		status = CAN_GetValue(pChannel, pParamater, charBuffer,
				(WORD) BufferLength);

		// Append Version Info To String Buffer
		if (status == PCAN_ERROR_OK)
			appendTextToJavaStringBuffer(env, Buffer, charBuffer);
	} else {
		//Call PCANBasic Function
		status = CAN_GetValue(pChannel, pParamater, &intBuffer,
				sizeof(intBuffer));

		//Parse int Result Into Java Integer To Maintain Reference
		cls = (*env)->FindClass(env, "peak/can/MutableInteger");
		mid = (*env)->GetMethodID(env, cls, "setValue", "(I)V");
		(*env)->CallVoidMethod(env, Buffer, mid, intBuffer);
	}

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Configures or sets a PCAN Channel value. </summary>
///
/// <remarks>
/// Parameters can be present or not according with the kind of Hardware (PCAN Channel) being
/// used. If a parameter is not available, a PCAN_ERROR_ILLPARAMTYPE error will be returned.
/// </remarks>
///
/// <param name="env">			[in,out] If non-null, the env. </param>
/// <param name="obj">			The object. </param>
/// <param name="Channel">		"The handle of a PCAN Channel". </param>
/// <param name="Parameter">	"The TPCANParameter parameter to set". </param>
/// <param name="Buffer">		"Buffer with the value to be set". </param>
/// <param name="BufferLength">	"Size in bytes of the buffer". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_SetValue(JNIEnv *env,
		jobject obj, jobject Channel, jobject Parameter, jobject Buffer,
		jint BufferLength) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;
	TPCANParameter pParamater;
	jclass cls;
	jmethodID mid;
	jstring toStringResult;
	const char* charBuffer;
	char charBufferMax[MAX_PATH];
	int intBuffer;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Parse Parameter
	pParamater
			= (TPCANParameter) ParseEnumValueFromJava(env, &Parameter, "()I");

	// Check Buffer Is Instance Of StringBuffer
	if ((*env)->IsInstanceOf(env, Buffer, (*env)->FindClass(env,
			"java/lang/StringBuffer")) == JNI_TRUE) {
		// Retrieves content from the Java StringBuffer object
		cls = (*env)->GetObjectClass(env, Buffer);
		mid = (*env)->GetMethodID(env, cls, "toString", "()Ljava/lang/String;");
		toStringResult = (jstring) (*env)->CallObjectMethod(env, Buffer, mid);
		// Convert from jstring to const char*
		charBuffer = (*env)->GetStringUTFChars(env, toStringResult, NULL);
		strcpy(charBufferMax, charBuffer);
		// Free Resource to avoid memory leak ...
		(*env)->ReleaseStringUTFChars(env, toStringResult, charBuffer);
		// Call PCANBasic Function
		status = CAN_SetValue(pChannel, pParamater, charBufferMax,
				(WORD) BufferLength);
	} else {
		// Parse int Result Into Java Integer To Maintain Reference
		cls = (*env)->GetObjectClass(env, Buffer);
		mid = (*env)->GetMethodID(env, cls, "getValue", "()I");
		intBuffer = (int) (*env)->CallIntMethod(env, Buffer, mid);

		//Call PCANBasic Function
		status = CAN_SetValue(pChannel, pParamater, &intBuffer,
				sizeof(intBuffer));
	}

	//Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Configures the reception filter. </summary>
///
/// <remarks>
/// The message filter will be expanded with every call to this function. If it is desired to
/// reset the filter, please use the CAN_SetParameter function.
/// </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
/// <param name="FromID">	"The lowest CAN ID to be received". </param>
/// <param name="ToID">		"The highest CAN ID to be received". </param>
/// <param name="Mode">		"Message type, Standard (11-bit identifier) or Extended (29-bit
/// 						identifier)". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_FilterMessages(
		JNIEnv *env, jobject obj, jobject Channel, jint FromID, jint ToID,
		jobject Mode) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;
	TPCANParameter pMode;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Parse Mode
	pMode = (TPCANMode) ParseEnumValueFromJava(env, &Mode, "()B");

	// Call PCANBasic Function
	status = CAN_FilterMessages(pChannel, FromID, ToID, pMode);

	// Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Transmits a CAN message. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">				[in,out] If non-null, the env. </param>
/// <param name="obj">				The object. </param>
/// <param name="Channel">			"The handle of a PCAN Channel". </param>
/// <param name="MessageBuffer">	"A TPCANMsg buffer with the message to be sent". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_Write(JNIEnv *env,
		jobject obj, jobject Channel, jobject MessageBuffer) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;
	TPCANMsg message;
	int i;
	jclass cls;
	jmethodID mid;
	jbyteArray javaByteArray;
	jbyte *nativeByteArray;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Verify MessageBuffer is not NULL
	if (MessageBuffer == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: MessageBuffer is null.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}

	//Gets class
	cls = (*env)->GetObjectClass(env, MessageBuffer);
	if (cls == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetObjectClass failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}

	//Gets Method ID
	mid = (*env)->GetMethodID(env, cls, "getID", "()I");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetMethodID('getID') failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}
	message.ID = (*env)->CallIntMethod(env, MessageBuffer, mid);

	//Gets Type
	mid = (*env)->GetMethodID(env, cls, "getType", "()B");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetMethodID('getType') failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}
	message.MSGTYPE = (*env)->CallByteMethod(env, MessageBuffer, mid);

	//Gets Length
	mid = (*env)->GetMethodID(env, cls, "getLength", "()B");
	if (mid == NULL) {
		ThrowExByName(
				env,
				"java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetMethodID('getLength') failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}
	message.LEN = (*env)->CallByteMethod(env, MessageBuffer, mid);

	//get Data
	mid = (*env)->GetMethodID(env, cls, "getData", "()[B");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetMethodID('getData') failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}
	javaByteArray = (jbyteArray) (*env)->CallObjectMethod(env, MessageBuffer,
			mid);
	if (javaByteArray == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: CallObjectMethod failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}

	//convert byte array to native byte array
	nativeByteArray = (*env)->GetByteArrayElements(env, javaByteArray, NULL);
	if (nativeByteArray == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"Java_peak_can_basic_PCANBasic_Write function: GetByteArrayElements failed.");
		ParseTPCANStatusToJava(env, PCAN_ERROR_UNKNOWN, &jStatus);
		return jStatus;
	}

	for (i = 0; i < (message.LEN); i++) {
		(message.DATA)[i] = nativeByteArray[i];
	}

	// Calls PCANBasic Function
	status = CAN_Write(pChannel, &message);

	// Parses Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Resets the receive and transmit queues of the PCAN Channel. </summary>
///
/// <remarks>	A reset of the CAN controller is not performed. </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_Reset(JNIEnv *env,
		jobject obj, jobject Channel) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Call PCANBasic Function
	status = CAN_Reset(pChannel);

	// Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Gets the current status of a PCAN Channel. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="obj">		The object. </param>
/// <param name="Channel">	"The handle of a PCAN Channel". </param>
///
/// <returns>	"A TPCANStatus error code". </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

JNIEXPORT jobject JNICALL Java_peak_can_basic_PCANBasic_GetStatus(JNIEnv *env,
		jobject obj, jobject Channel) {
	TPCANStatus status;
	jobject jStatus;
	TPCANHandle pChannel;

	// Parse Channel
	pChannel = (TPCANHandle) ParseEnumValueFromJava(env, &Channel, "()B");

	// Call PCANBasic Function
	status = CAN_GetStatus(pChannel);

	// Parse Result
	ParseTPCANStatusToJava(env, status, &jStatus);

	return jStatus;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	 Parse C/TPCANStatus to the corresponding Java/TPCANStatus. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in] If non-null, the env. </param>
/// <param name="source">	Source for the. </param>
/// <param name="target">	[in,out] If non-null, target for the. </param>
////////////////////////////////////////////////////////////////////////////////////////////////////

void ParseTPCANStatusToJava(JNIEnv *env, TPCANStatus source, jobject* target) {
	char* constantName = NULL;
	switch (source) {
	case PCAN_ERROR_OK:
		constantName = "PCAN_ERROR_OK";
		break;
	case PCAN_ERROR_XMTFULL:
		constantName = "PCAN_ERROR_XMTFULL";
		break;
	case PCAN_ERROR_OVERRUN:
		constantName = "PCAN_ERROR_OVERRUN";
		break;
	case PCAN_ERROR_BUSLIGHT:
		constantName = "PCAN_ERROR_BUSLIGHT";
		break;
	case PCAN_ERROR_BUSOFF:
		constantName = "PCAN_ERROR_BUSOFF";
		break;
	case PCAN_ERROR_ANYBUSERR:
		constantName = "PCAN_ERROR_ANYBUSERR";
		break;
	case PCAN_ERROR_QRCVEMPTY:
		constantName = "PCAN_ERROR_QRCVEMPTY";
		break;
	case PCAN_ERROR_QOVERRUN:
		constantName = "PCAN_ERROR_QOVERRUN";
		break;
	case PCAN_ERROR_QXMTFULL:
		constantName = "PCAN_ERROR_QXMTFULL";
		break;
	case PCAN_ERROR_REGTEST:
		constantName = "PCAN_ERROR_REGTEST";
		break;
	case PCAN_ERROR_RESOURCE:
		constantName = "PCAN_ERROR_RESOURCE";
		break;
	case PCAN_ERROR_ILLPARAMTYPE:
		constantName = "PCAN_ERROR_ILLPARAMTYPE";
		break;
	case PCAN_ERROR_ILLPARAMVAL:
		constantName = "PCAN_ERROR_ILLPARAMVAL";
		break;
	case PCAN_ERROR_ILLHANDLE:
		constantName = "PCAN_ERROR_ILLHANDLE";
		break;
	case PCAN_ERROR_INITIALIZE:
		constantName = "PCAN_ERROR_INITIALIZE";
		break;
	case PCAN_ERROR_NODRIVER:
		constantName = "PCAN_ERROR_NODRIVER";
		break;
	default:
		constantName = "PCAN_ERROR_UNKNOWN";
		break;
	}
	GetClassEnumValue(env, target, "peak/can/basic/TPCANStatus", constantName,
			"Lpeak/can/basic/TPCANStatus;");
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Parse tpcan handle to java. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in,out] If non-null, the env. </param>
/// <param name="source">	The Source . </param>
/// <param name="target">	[in,out] If non-null, the target. </param>
////////////////////////////////////////////////////////////////////////////////////////////////////

void ParseTPCANHandleToJava(JNIEnv *env, TPCANHandle source, jobject* target) {
	const char* constantName = NULL;
	switch (source) {
	case PCAN_NONEBUS:
		constantName = "PCAN_NONEBUS";
		break;
	case PCAN_ISABUS1:
		constantName = "PCAN_ISABUS1";
		break;
	case PCAN_ISABUS2:
		constantName = "PCAN_ISABUS2";
		break;
	case PCAN_ISABUS3:
		constantName = "PCAN_ISABUS3";
		break;
	case PCAN_ISABUS4:
		constantName = "PCAN_ISABUS4";
		break;
	case PCAN_ISABUS5:
		constantName = "PCAN_ISABUS5";
		break;
	case PCAN_ISABUS6:
		constantName = "PCAN_ISABUS6";
		break;
	case PCAN_ISABUS7:
		constantName = "PCAN_ISABUS7";
		break;
	case PCAN_ISABUS8:
		constantName = "PCAN_ISABUS8";
		break;
	case PCAN_DNGBUS1:
		constantName = "PCAN_DNGBUS1";
		break;
	case PCAN_PCIBUS1:
		constantName = "PCAN_PCIBUS1";
		break;
	case PCAN_PCIBUS2:
		constantName = "PCAN_PCIBUS2";
		break;
	case PCAN_PCIBUS3:
		constantName = "PCAN_PCIBUS3";
		break;
	case PCAN_PCIBUS4:
		constantName = "PCAN_PCIBUS4";
		break;
	case PCAN_PCIBUS5:
		constantName = "PCAN_PCIBUS5";
		break;
	case PCAN_PCIBUS6:
		constantName = "PCAN_PCIBUS6";
		break;
	case PCAN_PCIBUS7:
		constantName = "PCAN_PCIBUS7";
		break;
	case PCAN_PCIBUS8:
		constantName = "PCAN_PCIBUS8";
		break;
	case PCAN_USBBUS1:
		constantName = "PCAN_USBBUS1";
		break;
	case PCAN_USBBUS2:
		constantName = "PCAN_USBBUS2";
		break;
	case PCAN_USBBUS3:
		constantName = "PCAN_USBBUS3";
		break;
	case PCAN_USBBUS4:
		constantName = "PCAN_USBBUS4";
		break;
	case PCAN_USBBUS5:
		constantName = "PCAN_USBBUS5";
		break;
	case PCAN_USBBUS6:
		constantName = "PCAN_USBBUS6";
		break;
	case PCAN_USBBUS7:
		constantName = "PCAN_USBBUS7";
		break;
	case PCAN_USBBUS8:
		constantName = "PCAN_USBBUS8";
		break;
	case PCAN_PCCBUS1:
		constantName = "PCAN_PCCBUS1";
		break;
	case PCAN_PCCBUS2:
		constantName = "PCAN_PCCBUS2";
		break;
	}
	GetClassEnumValue(env, target, "peak/can/basic/TPCANHandle", constantName,
			"Lpeak/can/basic/TPCANHandle;");
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Parse C/TPCANStatus to the corresponding Java/TPCANStatus. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in] If non-null, the env. </param>
/// <param name="source">	[in] If non-null, source for the. </param>
/// <param name="type">		The type. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

int ParseEnumValueFromJava(JNIEnv *env, jobject* source, const char *type) {
	jclass cls;
	jmethodID mid;
	cls = (*env)->GetObjectClass(env, *source);
	if (cls == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"ParseEnumValueFromJava: GetObjectClass failed.");
		return -1;
	}

	// set millis	
	mid = (*env)->GetMethodID(env, cls, "getValue", type);
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"ParseEnumValueFromJava: GetMethodID failed.");
		return -1;
	}
	return (*env)->CallIntMethod(env, *source, mid);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Appends a string at the end of a Java StringBuffer. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">			[in] If non-null, the env. </param>
/// <param name="stringBuffer">	Buffer for string data. </param>
/// <param name="strValue">		[in] If non-null, the string value. </param>
////////////////////////////////////////////////////////////////////////////////////////////////////

void appendTextToJavaStringBuffer(JNIEnv *env, jobject stringBuffer,
		char* strValue) {
	jclass clazz;
	jmethodID mid;

	// Obtain the Java StringBuffer class handle that corresponds to the 
	// Java StringBuffer object handle.
	clazz = (*env)->GetObjectClass(env, stringBuffer);
	// Obtain the method ID for the StringBuffer append method which takes
	// a StringBuffer object reference argument and returns a String object
	// reference.
	mid = (*env)->GetMethodID(env, clazz, "append",
			"(Ljava/lang/String;)Ljava/lang/StringBuffer;");
	// If this method does not exist then return.
	if (mid == 0)
		return;
	else {
		// Create a new Java String object and populate it with the environment
		// variable value.  Obtain a handle to this object.
		jstring str = (*env)->NewStringUTF(env, strValue);
		// Call the StringBuffer object's append method passing the handle to 
		// the new Java String object.
		(*env)->CallObjectMethod(env, stringBuffer, mid, str);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Parse can message to java. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="env">		[in] If non-null, the env. </param>
/// <param name="mymsg">	[in] If non-null, the mymsg. </param>
/// <param name="msg">		[out] If non-null, the message. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

// Copy TPCANMsg object to a Java CANMessage object
int ParseCanMessageToJava(JNIEnv *env, TPCANMsg* mymsg, jobject* msg) {
	jclass cls;
	jmethodID mid;
	jbyteArray byteArray;

	// Get Java class
	cls = (*env)->GetObjectClass(env, *msg);
	if (cls == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetObjectClass for TPCANMsg failed.");
		return -1;
	}

	// Set ID	
	mid = (*env)->GetMethodID(env, cls, "setID", "(I)V");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetMethodID 'setID' for TPCANMsg failed.");
		return -1;
	}
	(*env)->CallVoidMethod(env, *msg, mid, (jint) (*mymsg).ID);

	// Set Type
	mid = (*env)->GetMethodID(env, cls, "setType", "(B)V");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetMethodID 'setType' for TPCANMsg failed.");
		return -1;
	}
	(*env)->CallVoidMethod(env, *msg, mid, (jbyte) (*mymsg).MSGTYPE);

	// Set Length
	mid = (*env)->GetMethodID(env, cls, "setLength", "(B)V");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetMethodID 'setLength' for TPCANMsg failed.");
		return -1;
	}
	(*env)->CallVoidMethod(env, *msg, mid, (jbyte) (*mymsg).LEN);

	// Create java byte array
	byteArray = (*env)->NewByteArray(env, (*mymsg).LEN);
	if (byteArray == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"NewbyteArray failed.");
		return -1;
	}
	(*env)->SetByteArrayRegion(env, byteArray, 0,
			(jbyte) (signed char) (*mymsg).LEN, (signed char*) (*mymsg).DATA);

	// Set Data
	mid = (*env)->GetMethodID(env, cls, "setData", "([BB)V");
	if (mid == NULL) {
		ThrowExByName(env, "java/lang/NullPointerException",
				"GetMethodID 'setData' for TPCANMsg failed.");
		(*env)->DeleteLocalRef(env, byteArray);
		return -1;
	}
	(*env)->CallVoidMethod(env, *msg, mid, byteArray, (jbyte) (*mymsg).LEN);

	// Delete reference
	(*env)->DeleteLocalRef(env, byteArray);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/// <summary>	Thread function which use the Received-Event. </summary>
///
/// <remarks>	 </remarks>
///
/// <param name="data">	[in] If non-null, the thread parameters. </param>
///
/// <returns>	. </returns>
////////////////////////////////////////////////////////////////////////////////////////////////////

void* CANReadThreadFunc(void *data) {
	TPCANStatus status;
	JNIEnv *m_env;
	jclass cls;
	jmethodID mid;
	CANReadThreadFuncParameter* paramater;
	jobject jHandle;
	fd_set Fds;
	struct timeval tv;
	int fd;



	//Get Parameter
	paramater = (CANReadThreadFuncParameter*) data;

	//Call PCANBasic Function
	status = CAN_GetValue(paramater->channel, PCAN_RECEIVE_EVENT, &fd,
			sizeof(int));

	// We attach the current thread to the loaded JVM
	(*m_vm)->AttachCurrentThread(m_vm, (void**) &m_env, NULL );
	// Get RcvEventMediator class
	cls = (*m_env)->FindClass(m_env, "peak/can/basic/RcvEventDispatcher");
	// Get static method dispatchRcvEvent
	mid = (*m_env)->GetStaticMethodID(m_env, cls, "dispatchRcvEvent",
			"(Lpeak/can/basic/TPCANHandle;)V");


	//Parse Handle
	ParseTPCANHandleToJava(m_env, paramater->channel, &jHandle);

	// set fds
	FD_ZERO(&Fds);
	FD_SET(fd, &Fds);
    tv.tv_sec = 0;
    tv.tv_usec = 500000;


	//Process
	while (!paramater->terminating)
		if (select(fd + 1, &Fds, NULL, NULL, NULL) > 0) {
			// Call dispatchRcvEvent to notify java application
			(*m_env)->CallStaticVoidMethod(m_env, cls, mid, jHandle);
		}
	return 0;
}

